package be.iminds.ilabt.jfed.ssh_terminal_tool;

import be.iminds.ilabt.jfed.gui_preferences.HLPreferenceKey;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.gui_preferences.JFedHLPreferences;
import be.iminds.ilabt.jfed.preferences.ProxyPreferencesManager;
import be.iminds.ilabt.jfed.util.common.OSDetector;
import be.iminds.ilabt.jfed.util.lib.LocalForwardUtil;
import javafx.application.Platform;
import javafx.scene.control.*;
import javafx.scene.layout.Region;
import javafx.scene.layout.VBox;
import javafx.stage.Stage;
import org.controlsfx.dialog.ExceptionDialog;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.regex.Pattern;

@Singleton
public class ExternalRdpApp {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ExternalRdpApp.class);

    private final GeniUserProvider geniUserProvider;
    private final JFedHLPreferences jFedPreferences;
    private final ProxyPreferencesManager proxyPreferencesManager;
    private final LocalForwardUtil localForwardUtil;

    @Inject
    public ExternalRdpApp(GeniUserProvider geniUserProvider,
                          ProxyPreferencesManager proxyPreferencesManager,
                          JFedHLPreferences jFedPreferences,
                          LocalForwardUtil localForwardUtil) {
        this.geniUserProvider = geniUserProvider;
        this.proxyPreferencesManager = proxyPreferencesManager;
        this.jFedPreferences = jFedPreferences;
        this.localForwardUtil = localForwardUtil;
    }

    public boolean isConfigured() {
        //TODO rewrite
        LOG.debug("ExternalRdpApp.isConfigured");
        if (OSDetector.os == OSDetector.OS.WIN) {
            File puttyDir = jFedPreferences.getFile(HLPreferenceKey.PREF_PUTTY_DIRECTORY);
            LOG.debug("ExternalRdpApp.isConfigured WIN puttyDir=" + puttyDir);
            if (puttyDir == null) return false;
            boolean valid = JFedHLPreferences.isValidPuttyDir(puttyDir);
            LOG.debug("ExternalRdpApp.isConfigured WIN puttyDir=" + puttyDir + " valid=" + valid);
            return valid;
        }

        if (OSDetector.os == OSDetector.OS.UNIX) {
            return jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL) != null;
        }

        if (OSDetector.os == OSDetector.OS.MAC) {
            return jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_UNIX_TERMINAL) != null;
        }

        return true;
    }

    public void launch(@Nullable String targetServerUserName, @Nonnull String hostName, int port,
                       @Nullable String password,
                       @Nullable JFedConnection.ProxyInfo customProxyInfo) {
        final GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();

        //This will set the proxyInfo to what is needed, based on if proxy exists, preferences and connectivity data.
        JFedConnection.SshProxyInfo jFedProxyInfo = geniUser.getUserAuthorityServer() == null ? null :
                proxyPreferencesManager.getSshProxySettings(geniUser.getUserAuthorityServer().getTestbed().getProxies(), geniUser);

        JFedConnection.SshProxyInfo usedProxyInfo = jFedProxyInfo;
        if (customProxyInfo != null) {
            usedProxyInfo = (JFedConnection.SshProxyInfo) customProxyInfo;
        }

        if (usedProxyInfo == null)
            LOG.info("ExternalRdpApp: No proxy will be used.");
        else {
            LOG.info("ExternalRdpApp: Proxy will be used: " + usedProxyInfo);

            //setup connection to server over the proxy, forwarding to a local port
            int forwardPort = 0;
            try {
                forwardPort = localForwardUtil.forwardOverSshProxy(usedProxyInfo, hostName, port);
            } catch (IOException e) {
                showSshPortForwardingFailureDialog(e, null);
                return;
                //throw new RuntimeException("Could not setup SSH port forwarding", e);
            }

            //Now continue, but make the connection to localhost instead
            hostName = "localhost";
            port = forwardPort;
        }

        if (OSDetector.os == OSDetector.OS.WIN) {
            launchWindowsRemoteDesktop(
                    targetServerUserName,
                    hostName,
                    port,
                    password);
            return;
        }

//        if (OSDetector.os == OSDetector.OS.UNIX || OSDetector.os == OSDetector.OS.MAC) {
        launchShellRdpApp(
                targetServerUserName,
                hostName,
                port,
                password);
        return;
//        }
//
//        throw new RuntimeException("No Rdp App known for OS \"" + OSDetector.os + "\"");
    }

    private void showSshPortForwardingFailureDialog(Exception e, Stage stage) {
        ExceptionDialog exceptionDialog = new ExceptionDialog(e);
        if (stage != null)
            exceptionDialog.initOwner(stage);

        String headerText = "SSH Port Forwarding Failure";
        String contentText = "jFed encountered an exception while setting up SSH port forwarding to be used " +
                "for the Remote Desktop Session." +
                "This is likely caused by a connectivity issue or an error at the server. Consult the stacktrace or the " +
                "logs for more details." +
                "Please send the jFed developers a bug report if this is a bug in jFed.";

//        VBox contentPane = new VBox(10);
//        contentPane.setPrefWidth(600);
//        Label sorryLabel = new Label(headerText); //"Connectivity Issue";
//        sorryLabel.setFont(Font.font(Font.getDefault().getFamily(), FontWeight.BOLD, Font.getDefault().getSize()));
//
//        Label infoLabel = new Label(contentText);
//        infoLabel.setWrapText(true);

//        contentPane.getChildren().addAll(sorryLabel, infoLabel);
//        exceptionDialog.getDialogPane().setContent(contentPane);

        exceptionDialog.setHeaderText(headerText);
        exceptionDialog.setContentText(contentText);

        exceptionDialog.show();
//        Platform.runLater(() -> {
//            exceptionDialog.showAndWait();
//        });
    }

    private void showWindowsRemoteDesktopDetails(@Nullable String targetServerUserName,
                                                 @Nonnull String hostName,
                                                 int port,
                                                 @Nullable String password) {
        VBox pane = new VBox();
        pane.setPrefWidth(750);
        pane.setSpacing(5);

        String text = String.format("jFed is not configured to open an RDP application." +
                        "\n" +
                        "\nYou can manually open a windows Remote Desktop connection to %s:%d\n",
                hostName, port);
        if (targetServerUserName != null || password != null)
            text += "You can login ";
        if (targetServerUserName != null)
            text += String.format(" with username \"%s\"", targetServerUserName);
        if (password != null)
            text += String.format(" using the password \"%s\"", password);

        Label textLabel = new Label(text);
        textLabel.setPrefHeight(75);

        textLabel.setWrapText(true);
        pane.getChildren().add(textLabel);

        DialogPane dialogPane = new DialogPane();
        dialogPane.setContent(pane);
        dialogPane.getButtonTypes().setAll(ButtonType.OK);


        Dialog<ButtonType> dialog = new Dialog<>();
        dialog.setTitle("Remote Desktop Info");
        dialog.setDialogPane(dialogPane);

        dialog.showAndWait();
    }

    private void launchWindowsRemoteDesktop(@Nullable String username,
                                            @Nonnull String hostName,
                                            int port,
                                            @Nullable String password) {
        String rdpCommandTemplate = jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_RDP);
        if (rdpCommandTemplate == null || rdpCommandTemplate.trim().isEmpty()) {
            showWindowsRemoteDesktopDetails(username, hostName, port, password);
            return;
        }

        launchExternalRdpApp(rdpCommandTemplate, hostName, port, password, rdpCommandTemplate);
    }


    private void launchShellRdpApp(@Nullable String username,
                                   @Nonnull String hostName,
                                   int port,
                                   @Nullable String password) {
        String rdpCommandTemplate = jFedPreferences.getString(HLPreferenceKey.PREF_COMMAND_RDP);
        if (rdpCommandTemplate == null || rdpCommandTemplate.trim().isEmpty()) {
            showWindowsRemoteDesktopDetails(username, hostName, port, password);
            return;
        }

        launchExternalRdpApp(username, hostName, port, password, rdpCommandTemplate);
    }

    private void launchExternalRdpApp(@Nullable String username,
                                      @Nonnull String hostName,
                                      int port,
                                      @Nullable String password,
                                      String rdpCommandTemplate) {
        Runnable processOutputMonitor = () -> {
            try {
                //example: rdesktop -p <pass> -u <username> <server>:<port>

                LinkedList<String> commandList = new LinkedList<>(Arrays.asList(rdpCommandTemplate.split(" ")));
                for (ListIterator<String> it = commandList.listIterator(); it.hasNext();) {
                    String parameter = it.next();
                    if (parameter.equals("$P")) {
                        if (password == null) {
                            //remove parameter, AND previous parameter
                            it.remove();
                            it.previous();
                            it.remove();
                            continue;
                        } else {
                            it.set(password);
                        }
                    }
                    if (parameter.equals("$u")) {
                        if (username == null) {
                            //remove parameter, AND previous parameter
                            it.remove();
                            it.previous();
                            it.remove();
                            continue;
                        } else {
                            it.set(username);
                        }
                    }
                    String replacement = parameter;
                    if (parameter.contains("$h")) {
                        replacement = replacement.replaceAll(Pattern.quote("$h"), hostName);
                        it.set(replacement);
                    }
                    if (parameter.contains("$p")) {
                        replacement = replacement.replaceAll(Pattern.quote("$p"), port+"");
                        it.set(replacement);
                    }
                }

                //unix
                LOG.debug("ExternalRdpApp Thread: Starting terminal using command (list): " + commandList);

                ProcessBuilder pb = new ProcessBuilder(commandList);
                pb = pb.redirectErrorStream(true);

                Instant startInstant = Instant.now();

                Process p;
                try {
                    p = pb.start();
                    BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));

                    String line;
                    while ((line = input.readLine()) != null) {
                        LOG.debug("ExternalRdpApp Thread: External RdpApp stdout+stderr: " + line);
                    }
                    //note: input might close before terminal stops.
                    input.close();
                    try {
                        p.waitFor();
                    } catch (InterruptedException e) {
                        LOG.warn("ExternalRdpApp Thread: processOutputMonitor thread interrupted while waiting for terminal to stop", e);
                    }
                } catch (Exception e) {
                    LOG.error("Error launching RDP client", e);
                    p = null;
                }
                Instant stopInstant = Instant.now();

                if (p != null)
                    LOG.info("ExternalRdpApp Thread: Terminal command exited with exit value: " + p.exitValue());

                if (stopInstant.isBefore(startInstant.plus(2, ChronoUnit.SECONDS))) {
                    //Command exited too soon.
                    Platform.runLater(() -> {
                        Alert alert = new Alert(Alert.AlertType.ERROR);
//                    alert.initOwner(dialogOwner);
                        alert.setContentText("Problem launching Windows Remote Desktop Application:\n" +
                                "The application was started, but exited very soon afterwards.\n\n" +
                                "Check the command in the preferences, or the jFed logs (CTRL-F10)."
                        );
                        alert.setTitle("RDP App Problem");
                        alert.setResizable(true);
                        alert.getDialogPane().setMinHeight(Region.USE_PREF_SIZE);
                        alert.getButtonTypes().setAll(ButtonType.CLOSE);
                        alert.showAndWait();
                    });
                }
            } catch (Exception e) {
                LOG.error("Unhandled Exception in ExternalRdpApp Thread.", e);
            } catch (AssertionError e) {
                LOG.error("Unhandled AssertionError in ExternalRdpApp Thread.", e);
            }
        };
        try {
            Thread t = new Thread(processOutputMonitor);
            t.setName("ExternalRdpApp");
            t.setDaemon(true); //=> this application may exit before rdp app stops
            t.start();
        } catch (AssertionError e) {
            LOG.error("AssertionError starting ExternalRdpApp Thread", e);
        } catch (Exception e) {
            LOG.error("Exception starting ExternalRdpApp Thread", e);
        } catch (Error e) {
            //ok to catch, because thrown at end
            //noinspection finally
            try {
                LOG.error("Error starting ExternalRdpApp Thread", e);
            } finally {
                throw e;
            }
        }
    }
}

